# Copyrights 2010-2013 Pierre Chanial
# All rights reserved
#

import astropy.wcs
import numpy as np
import scipy.interpolate as interp
from astropy.io import fits as pyfits
from astropy.wcs import WCS

from pyoperators import Operator
from pyoperators.flags import inplace, real, square
from pyoperators.utils import isscalarlike, operation_assignment

from . import _flib as flib

__all__ = [
    'angle_lonlat',
    'barycenter_lonlat',
    'create_fitsheader',
    'create_fitsheader_for',
    'fitsheader2shape',
    'str2fitsheader',
    'DistortionOperator',
    'RotationBoresightEquatorialOperator',
    'WCSToPixelOperator',
    'WCSToWorldOperator',
]

# Currently, astropy wcs transform only handle input of ndim equal to 2
# See https://github.com/astropy/astropy/issues/976
ASTROPY_WCS_NDIM_IS_2 = True


def angle_lonlat(lon1, lat1, lon2=None, lat2=None):
    """
    Returns the angle between vectors on the celestial sphere in degrees.

    Parameters
    ----------
    lon1, lon2 : array of numbers
        longitude in degrees
    lat1, lat2 : array of numbers
        latitude in degrees

    Example
    -------
    >>> ra1, dec1 = 30, 70
    >>> ra2, dec2 = 30.5, 70.5
    >>> angle_lonlat((ra1, dec1), (ra2, dec2))
    >>> angle_lonlat(ra1, dec1, ra2, dec2)

    """
    if lon2 is None and lat2 is None:
        lon2, lat2 = lat1
        lon1, lat1 = lon1
    lon1 = np.asarray(lon1, float)
    lat1 = np.asarray(lat1, float)
    lon2 = np.asarray(lon2, float)
    lat2 = np.asarray(lat2, float)
    angle = flib.wcsutils.angle_lonlat(
        lon1.ravel(), lat1.ravel(), lon2.ravel(), lat2.ravel()
    )
    if all(_.ndim == 0 for _ in (lon1, lat1, lon2, lat2)):
        angle = angle[0]
    return angle


def barycenter_lonlat(lon, lat):
    """
    Returns the barycenter of vectors on the celestial sphere.

    Parameters
    ----------
    lon : array of numbers
        Longitude in degrees.
    lat : array of numbers
        Latitude in degrees.
    """
    lon = np.atleast_1d(np.asarray(lon, float)).ravel()
    lat = np.atleast_1d(np.asarray(lat, float)).ravel()
    return flib.wcsutils.barycenter_lonlat(lon, lat)


def combine_fitsheader(headers, cdelt=None, pa=None):
    """
    Returns a FITS header which encompasses all input headers.

    Parameters
    ----------
    headers : list of pyfits.Header
        The headers of the maps that must be contained in the map of the
        returned header.
    cdelt : float or 2 element array
        Physical increment at the reference pixel. Default is the lowest
        header cdelt.
    pa : float
        Position angle of the Y=AXIS2 axis (=-CROTA2). Default is the mean
        header pa.
    """

    if not isinstance(headers, (list, tuple)):
        headers = (headers,)

    if len(headers) == 1 and cdelt is None and pa is None:
        return headers[0]

    # compute the world coordinates of the reference pixel of the combined map
    crval = barycenter_lonlat(
        [h['CRVAL1'] for h in headers], [h['CRVAL2'] for h in headers]
    )

    # compute the combined map cdelt & pa
    if cdelt is None or pa is None:
        cdeltpa = [get_cdelt_pa(h) for h in headers]

    if cdelt is None:
        cdelt = min(np.min(abs(cdelt_)) for cdelt_, pa_ in cdeltpa)

    if pa is None:
        pa = flib.wcsutils.mean_degrees(np.array([pa_ for cdelt_, pa_ in cdeltpa]))

    # first, create the combined header with a single pixel centred on the
    # reference coordinates, with correct orientation and cdelt
    header0 = create_fitsheader((1, 1), cdelt=cdelt, pa=pa, crpix=(1, 1), crval=crval)

    # then, 'enlarge' it to make it fit the edges of each header
    proj0 = WCS(header0)
    xy0 = []
    for h in headers:
        p = WCS(h)
        nx = h['NAXIS1']
        ny = h['NAXIS2']
        x = h['CRPIX1']
        y = h['CRPIX2']
        edges = p.wcs_pix2world(
            [0.5, x, nx + 0.5, 0.5, x, nx + 0.5, 0.5, x, nx + 0.5],
            [0.5, 0.5, 0.5, y, y, y, ny + 0.5, ny + 0.5, ny + 0.5],
            1,
        )
        xy0.append(proj0.wcs_world2pix(edges[0], edges[1], 1))

    xmin0 = np.round(min(min(c[0]) for c in xy0))
    xmax0 = np.round(max(max(c[0]) for c in xy0))
    ymin0 = np.round(min(min(c[1]) for c in xy0))
    ymax0 = np.round(max(max(c[1]) for c in xy0))

    header0['NAXIS1'] = int(xmax0 - xmin0 + 1)
    header0['NAXIS2'] = int(ymax0 - ymin0 + 1)
    header0['CRPIX1'] = 2 - xmin0
    header0['CRPIX2'] = 2 - ymin0

    return header0


def create_fitsheader(
    naxes=None,
    dtype=None,
    fromdata=None,
    extname=None,
    crval=(0, 0),
    crpix=None,
    cunit='deg',
    ctype=('RA---TAN', 'DEC--TAN'),
    cd=None,
    cdelt=None,
    pa=None,
    radesys='ICRS',
    equinox=None,
):
    """
    Helper to create a FITS header.

    Parameters
    ----------
    naxes : array, optional
        (NAXIS1,NAXIS2,...) tuple, which specifies the data dimensions.
        Note that the dimensions in FITS and ndarrays are reversed.
    dtype : data-type, optional
        Desired data type for the data stored in the FITS file
    extname : None or string
        if a string is specified ('' can be used), the returned header
        type will be an Image HDU (otherwise a Primary HDU).
    crval : 2 element array, optional
        Reference pixel values (FITS convention).
    crpix : 2 element array, optional
        Reference pixel (FITS convention).
    ctype : 2 element string array, optional
        Projection types.
    cunit : string or 2 element string array
        Units of the CD matrix (default is degrees/pixel).
    cd : 2 x 2 array
        FITS parameters
            CD1_1 CD1_2
            CD2_1 CD2_2
    cdelt : float or 2 element array
        Physical increment at the reference pixel.
    pa : float
        Position angle of the Y=AXIS2 axis (=-CROTA2).
    radesys : string
        Coordinate reference frame for the RA and DEC axis columns. Default
        is 'ICRS'.
    equinox : float, optional
        Reference equinox (for non-ICRS reference frame).

    Example
    -------
    >>> from pysimulators import Map
    >>> map = Map.ones((10, 100), unit='Jy/pixel')
    >>> ny, nx = map.shape
    >>> map.header = create_fitsheader((nx, ny), cd=[[-1,0], [0,1]])

    """
    if naxes is not None and fromdata is not None or naxes is None and fromdata is None:
        raise ValueError("Either keyword 'naxes' or 'fromdata' must be specified.")

    if fromdata is None:
        naxes = np.array(naxes, dtype=int, ndmin=1)
        naxes = tuple(naxes)
        if len(naxes) > 8:
            raise ValueError('First argument is naxes=(NAXIS1,NAXIS2,...)')
        if dtype is not None:
            dtype = np.dtype(dtype)
            typename = dtype.name
        else:
            typename = 'float64'
    else:
        print(
            "The keyword 'from_data' in create_fitsheader is deprecated and "
            'will be removed in an upcoming release. Use the function create'
            '_fitsheader_for instead.'
        )
        array = np.asarray(fromdata)
        naxes = tuple(reversed(array.shape))
        if dtype is not None:
            array = array.astype(dtype)
        if array.dtype.itemsize == 1:
            typename = 'uint8'
        elif array.dtype.names is not None:
            typename = None
        else:
            typename = array.dtype.name

    numaxis = len(naxes)

    header = pyfits.Header()
    if extname is None:
        header['simple'] = True  # primary header
    else:
        header['xtension'] = ('IMAGE', 'Image extension')
    if typename is not None:
        header['bitpix'] = (pyfits.DTYPE2BITPIX[typename], 'Array data type')
    header['naxis'] = (numaxis, 'Number of array dimensions')
    for dim in range(numaxis):
        header['naxis' + str(dim + 1)] = naxes[dim]
    if extname is None:
        header['extend'] = True
    else:
        header['pcount'] = (0, 'Number of parameters')
        header['gcount'] = (1, 'Number of groups')
        header['extname'] = extname

    if cd is not None:
        cd = np.asarray(cd, dtype=float)
        if cd.shape != (2, 2):
            raise ValueError('The CD matrix is not a 2x2 matrix.')
    else:
        if cdelt is None:
            return header
        cdelt = np.array(cdelt, float)
        if isscalarlike(cdelt):
            if (
                ctype is None
                or len(ctype) == 2
                and ctype[0] == 'RA---TAN'
                and ctype[1] == 'DEC--TAN'
            ):
                cdelt = (-cdelt, cdelt)
            else:
                cdelt = (cdelt, cdelt)
        if pa is None:
            pa = 0.0
        theta = np.deg2rad(-pa)
    cd = np.diag(cdelt).dot(
        np.array([[np.cos(theta), np.sin(theta)], [-np.sin(theta), np.cos(theta)]])
    )

    crval = np.asarray(crval, float)
    if crval.size != 2:
        raise ValueError('CRVAL does not have two elements.')

    if crpix is None:
        crpix = (np.array(naxes) + 1) / 2
    else:
        crpix = np.asarray(crpix, float)
    if crpix.size != 2:
        raise ValueError('CRPIX does not have two elements.')

    if len(ctype) != 2:
        raise ValueError('CTYPE does not have two elements.')

    if isscalarlike(cunit):
        cunit = (cunit, cunit)
    if len(cunit) != 2:
        raise ValueError('CUNIT does not have two elements.')

    header['crval1'] = crval[0]
    header['crval2'] = crval[1]
    header['crpix1'] = crpix[0]
    header['crpix2'] = crpix[1]
    header['cd1_1'] = cd[0, 0]
    header['cd2_1'] = cd[1, 0]
    header['cd1_2'] = cd[0, 1]
    header['cd2_2'] = cd[1, 1]
    header['ctype1'] = ctype[0]
    header['ctype2'] = ctype[1]
    header['cunit1'] = cunit[0]
    header['cunit2'] = cunit[1]

    if ctype[0][:3] == 'RA-' and ctype[1][:3] == 'DEC':
        header['radesys'] = radesys.upper()

    if equinox is not None:
        header['equinox'] = float(equinox)

    return header


def create_fitsheader_for(
    data,
    extname=None,
    crval=(0, 0),
    crpix=None,
    ctype=('RA---TAN', 'DEC--TAN'),
    cunit='deg',
    cd=None,
    cdelt=None,
    pa=None,
    radesys='ICRS',
    equinox=None,
):
    """
    Helper to create a FITS header from an ndarray.

    Parameters
    ----------
    data : array_like or None
        The array from which the dimensions and type will be extracted. Note
        that following the FITS convention, the dimension along X is the
        second value of the array shape and that the dimension along the
        Y axis is the first one.
    extname : None or string
        if a string is specified ('' can be used), the returned header
        type will be an Image HDU (otherwise a Primary HDU).
    crval : 2 element array, optional
        Reference pixel values (FITS convention).
    crpix : 2 element array, optional
        Reference pixel (FITS convention).
    ctype : 2 element string array, optional
        Projection types.
    cunit : string or 2 element string array
        Units of the CD matrix (default is degrees/pixel).
    cd : 2 x 2 array
        FITS parameters
            CD1_1 CD1_2
            CD2_1 CD2_2
    cdelt : float or 2 element array
        Physical increment at the reference pixel.
    pa : float
        Position angle of the Y=AXIS2 axis (=-CROTA2).
    radesys : string
        Coordinate reference frame for the RA and DEC axis columns. Default
        is 'ICRS'.
    equinox : float, optional
        Reference equinox (for non-ICRS reference frame).

    Example
    -------
    >>> from pysimulators import Map
    >>> map = Map.ones((10,100), unit='Jy/pixel')
    >>> map.header = create_fitsheader_for(map, cd=[[-1,0], [0,1]])

    >>> primary_header = create_fitsheader_for(None)
    >>> secondary_header = create_fitsheader_for(None, extname='secondary')

    """
    if data is None:
        naxes = ()
        dtype = np.int32
    else:
        data = np.array(data, ndmin=1)
        naxes = data.shape[::-1]
        if data.dtype.itemsize == 1:
            data = data.view('uint8')
        dtype = data.dtype
    return create_fitsheader(
        naxes,
        dtype=dtype,
        extname=extname,
        crval=crval,
        crpix=crpix,
        ctype=ctype,
        cunit=cunit,
        cd=cd,
        cdelt=cdelt,
        pa=pa,
        radesys=radesys,
        equinox=equinox,
    )


def fitsheader2shape(header):
    ndim = header['NAXIS']
    return tuple(header[f'NAXIS{i}'] for i in range(ndim, 0, -1))


def get_cdelt_pa(header):
    """
    Extract scale and rotation from a FITS header.

    Parameter
    ---------
    header : pyfits Header
        The FITS header

    Returns
    -------
    cdelt : 2 element array
        Physical increment at the reference pixel.
    pa : float
        Position angle of the Y=AXIS2 axis (=-CROTA2).

    """
    try:
        cd = np.array(
            [[header['CD1_1'], header['CD1_2']], [header['CD2_1'], header['CD2_2']]],
            float,
        )
    except KeyError:
        if any(k not in header for k in ('CDELT1', 'CDELT2', 'CROTA2')):
            raise KeyError('Header has no astrometry.')
        return (np.array([header['CDELT1'], header['CDELT2']]), -header['CROTA2'])

    det = np.linalg.det(cd)
    sgn = det / np.abs(det)
    cdelt = np.array(
        [
            sgn * np.sqrt(cd[0, 0] ** 2 + cd[0, 1] ** 2),
            np.sqrt(cd[1, 1] ** 2 + cd[1, 0] ** 2),
        ]
    )
    pa = -np.arctan2(-cd[1, 0], cd[1, 1])
    return cdelt, np.rad2deg(pa)


def has_wcs(header):
    """
    Returns True is the input FITS header has a defined World Coordinate
    System.
    """

    required = 'CRPIX,CRVAL,CTYPE'.split(',')
    keywords = np.concatenate(
        [
            (lambda i: [r + str(i + 1) for r in required])(i)
            for i in range(header['NAXIS'])
        ]
    )
    return all([k in header for k in keywords])


def mean_degrees(array):
    """
    Returns the mean value of an array of values in degrees, by taking into
    account the discrepancy at 0 degree
    """
    return flib.wcsutils.mean_degrees(np.asarray(array, dtype=float).ravel())


def minmax_degrees(array):
    """
    Returns the minimum and maximum value of an array of values in degrees,
    by taking into account the discrepancy at 0 degree.
    """
    return flib.wcsutils.minmax_degrees(np.asarray(array, dtype=float).ravel())


def str2fitsheader(string):
    """
    Convert a string into a pyfits.Header object

    All cards are extracted from the input string until the END keyword is
    reached.
    """
    header = pyfits.Header()
    cards = header.ascard()
    iline = 0
    while iline * 80 < len(string):
        line = string[iline * 80 : (iline + 1) * 80]
        if line[0:3] == 'END':
            break
        cards.append(pyfits.Card().fromstring(line))
        iline += 1
    return header


@real
@square
class DistortionOperator(Operator):
    """
    Distortion operator.

    The interpolation is performed using the Clough-Tocher method.
    Interpolation of points outside the object plane coordinates convex hull
    will return NaN.

    Parameters
    ----------
    xin : array of shape (npoints, ndims)
       Coordinates in the object plane.
    xout : array of shape (npoints, ndims)
       Coordinates in the image plane.

    """

    def __init__(self, xin, xout, **keywords):
        xin = np.asarray(xin)
        xout = np.asarray(xout)
        if xin.shape[-1] != 2:
            raise ValueError(
                'The shape of the object plane coordinates should'
                ' be (npoints,ndims), where ndims is only impleme'
                'nted for 2.'
            )
        if xin.shape != xout.shape:
            raise ValueError(
                'The object and image coordinates do not have the' ' same shape.'
            )

        keywords['dtype'] = float
        Operator.__init__(self, **keywords)
        self.interp0 = interp.CloughTocher2DInterpolator(xin, xout[..., 0])
        self.interp1 = interp.CloughTocher2DInterpolator(xin, xout[..., 1])
        self.set_rule('I', lambda s: DistortionOperator(xout, xin))

    def direct(self, input, output):
        output[..., 0] = self.interp0(input)
        output[..., 1] = self.interp1(input)


@real
class RotationBoresightEquatorialOperator(Operator):
    r"""
    Convert coordinates in the instrument object plane into celestial
    coordinates, assuming a pointing direction and a position angle.

            input[1] along declination if PA=0
       \    |
        \PA_|
         \/ |
          \ |
           \|_______ input[0] along -R.A. if PA=0

    The routine is not accurate at the poles.
    Input coordinates are in arc seconds, pointing and output are in degrees.

    """

    def __init__(self, pointing, **keywords):
        if any(k not in pointing.dtype.names for k in ('ra', 'dec', 'pa')):
            raise TypeError('The input is not an equatorial pointing.')
        self.pointing = pointing
        Operator.__init__(self, **keywords)

    def direct(self, x, out):
        s = (Ellipsis,) + (x.ndim - 1) * (None,)
        ra = self.pointing[s].ra
        dec = self.pointing[s].dec
        pa = np.deg2rad(self.pointing[s].pa)
        cospa = np.cos(pa) / 3600
        sinpa = np.sin(pa) / 3600
        out[..., 1] = dec + (-x[..., 0] * sinpa + x[..., 1] * cospa)
        out[..., 0] = ra + (x[..., 0] * cospa + x[..., 1] * sinpa) / np.cos(
            np.deg2rad(out[..., 1])
        )

    def reshapein(self, shape):
        return self.pointing.shape + shape


@real
@square
@inplace
class _WCSKapteynOperator(Operator):
    def __init__(self, wcs, **keywords):
        """
        wcs : FITS header or Kapteyn wcs.Projection instance
            Representation of the world coordinate system

        """
        import kapteyn.wcs as kwcs

        if isinstance(wcs, pyfits.Header):
            wcs = kwcs.Projection(wcs)
        if 'dtype' not in keywords:
            keywords['dtype'] = float
        wcs.allow_invalid = True
        Operator.__init__(self, **keywords)
        self.wcs = wcs

    def validatein(self, shapein):
        if len(shapein) == 0:
            raise ValueError('Invalid scalar input.')
        if shapein[-1] != 2:
            raise ValueError(f"Invalid last dimension: '{shapein[-1]}'.")


class WCSKapteynToPixelOperator(_WCSKapteynOperator):
    """
    Operator for WCS world-to-pixel transforms.

    Example
    -------
    >>> header = pyfits.open('myfile.fits')
    >>> w = WCSKapteynToPixelOperator(header)
    >>> ra0, dec0 = 190, 77
    >>> pix = w([ra0, dec0])

    """

    def __init__(self, wcs, **keywords):
        _WCSKapteynOperator.__init__(self, wcs, **keywords)
        self.set_rule('I', lambda s: WCSKapteynToWorldOperator(s.wcs))

    __init__.__doc__ = _WCSKapteynOperator.__init__.__doc__

    def direct(self, input, output, operation=operation_assignment):
        operation(output, self.wcs.topixel(input))


class WCSKapteynToWorldOperator(_WCSKapteynOperator):
    """
    Operator for WCS pixel-to-world transforms.

    Example
    -------
    >>> header = pyfits.open('myfile.fits')
    >>> w = WCSKapteynToWorldOperator(header)
    >>> x0, y0 = 34, 65.2
    >>> radec = w([x0, y0])

    """

    def __init__(self, wcs, **keywords):
        _WCSKapteynOperator.__init__(self, wcs, **keywords)
        self.set_rule('I', lambda s: WCSKapteynToPixelOperator(s.wcs))

    __init__.__doc__ = _WCSKapteynOperator.__init__.__doc__

    def direct(self, input, output, operation=operation_assignment):
        operation(output, self.wcs.toworld(input))


@real
@square
@inplace
class _WCSOperator(Operator):
    def __init__(self, wcs, origin=0, **keywords):
        """
        wcs : FITS header or astropy.wcs.WCS instance
            Representation of the world coordinate system
        origin : int, optional
            Indexing convention : 0 for 0-based indexing such as C or numpy),
            1 for FITS or Fortran 1-based indexing.

        """
        if not isinstance(wcs, astropy.wcs.WCS):
            wcs = astropy.wcs.WCS(wcs)
        if 'dtype' not in keywords:
            keywords['dtype'] = float
        Operator.__init__(self, **keywords)
        self.wcs = wcs
        self.origin = origin

    def validatein(self, shapein):
        if len(shapein) == 0:
            raise ValueError('Invalid scalar input.')
        if shapein[-1] != 2:
            raise ValueError(f"Invalid last dimension: '{shapein[-1]}'.")


class WCSToPixelOperator(_WCSOperator):
    """
    Operator for WCS world-to-pixel transforms.

    Example
    -------
    >>> header = pyfits.open('myfile.fits')
    >>> w = WCSToPixelOperator(header)
    >>> ra0, dec0 = 190, 77
    >>> pix = w([ra0, dec0])

    """

    def __init__(self, wcs, origin=0, **keywords):
        _WCSOperator.__init__(
            self, wcs, origin, classin=np.ndarray, classout=np.ndarray, **keywords
        )
        self.set_rule('I', lambda s: WCSToWorldOperator(s.wcs, s.origin))

    __init__.__doc__ = _WCSOperator.__init__.__doc__

    def direct(self, input, output, operation=operation_assignment):
        if ASTROPY_WCS_NDIM_IS_2:
            input = input.reshape((-1, 2))
            output = output.reshape((-1, 2))
        operation(output, self.wcs.wcs_world2pix(input, self.origin))


class WCSToWorldOperator(_WCSOperator):
    """
    Operator for WCS pixel-to-world transforms.

    Example
    -------
    >>> header = pyfits.open('myfile.fits')
    >>> w = WCSToWorldOperator(header)
    >>> x0, y0 = 34, 65.2
    >>> radec = w([x0, y0])

    """

    def __init__(self, wcs, origin=0, **keywords):
        _WCSOperator.__init__(
            self, wcs, origin, classin=np.ndarray, classout=np.ndarray, **keywords
        )
        self.set_rule('I', lambda s: WCSToPixelOperator(s.wcs, s.origin))

    __init__.__doc__ = _WCSOperator.__init__.__doc__

    def direct(self, input, output, operation=operation_assignment):
        if ASTROPY_WCS_NDIM_IS_2:
            input = input.reshape((-1, 2))
            output = output.reshape((-1, 2))
        operation(output, self.wcs.wcs_pix2world(input, self.origin))



import numpy as np
import pytest
from numpy.testing import assert_equal

from pyoperators.utils.testing import assert_same
from pysimulators import LayoutGridSquares, Quantity, SceneGrid, create_fitsheader
from pysimulators.geometry import create_grid_squares

ITYPES = np.int32, np.int64
FTYPES = np.float32, np.float64


@pytest.mark.parametrize(
    'center, expected',
    [
        ((0, 0), [[9, 10], [5, 6]]),
        (
            (-0.5, 0.5),
            [
                [0.25 * (8 + 9 + 12 + 13), 0.25 * (9 + 10 + 13 + 14)],
                [0.25 * (4 + 5 + 8 + 9), 0.25 * (5 + 6 + 9 + 10)],
            ],
        ),
    ],
)
@pytest.mark.parametrize('ncolmax', [0, 1, 4, 5])
@pytest.mark.parametrize('itype', ITYPES)
@pytest.mark.parametrize('ftype', FTYPES)
def test_spatial_integration(center, ncolmax, itype, ftype, expected):
    #  -------------
    #  |12|13|14|15|
    #  |--+--+--+--+
    #  | 8| 9|10|11|
    #  |--+--+--+--+
    #  | 4| 5| 6| 7|
    #  |--+--+--+--+
    #  | 0| 1| 2| 3|
    #  -------------
    #  The plane center has world coordinates (0, 0)

    plane_shape = (4, 4)
    pixsize = 1  # um
    header = create_fitsheader(
        plane_shape,
        ctype=['X---CAR', 'Y---CAR'],
        cdelt=[pixsize, pixsize],
        crval=(0, 0),
        cunit=['um', 'um'],
    )
    plane = SceneGrid.fromfits(header)

    detectors = create_grid_squares((2, 2), Quantity(pixsize, 'um'), center=center)
    proj = plane.get_integration_operator(
        plane.topixel(detectors), ncolmax=ncolmax, dtype=ftype, dtype_index=itype
    )
    x = np.arange(len(plane)).reshape(plane_shape)
    y = proj(x)

    assert_equal(proj.matrix.dtype, ftype)
    assert_equal(proj.matrix.data.index.dtype, itype)
    assert_equal(proj.shapein, plane.shape)
    assert_equal(proj.shapeout, detectors.shape[:-2])
    assert_equal(proj.matrix.shape, (4, len(plane)))
    expected_min_ncolmax = 1 if center == (0, 0) else 4
    assert_equal(proj.matrix.data.shape, (4, max(expected_min_ncolmax, ncolmax)))
    assert_equal(proj.min_ncolmax, expected_min_ncolmax)
    assert_same(y, expected)


@pytest.mark.parametrize('spacing', [0.5, 1, 2])
@pytest.mark.parametrize('angle', [0, 30])
def test_spatial_integration2(spacing, angle):
    shape_grid = (16, 16)
    filling_factor = 0.8

    origin = shape_grid * np.array(spacing) * np.sqrt(2) / 2
    grid = LayoutGridSquares(
        shape_grid,
        spacing,
        origin=origin,
        angle=angle,
        filling_factor=filling_factor,
    )
    shape_plane = np.ceil(shape_grid * np.array(spacing) * np.sqrt(2))
    scene = SceneGrid(shape_plane)
    proj = scene.get_integration_operator(grid.vertex)
    assert not proj.outside
    assert_same(
        np.sum(proj.matrix.data.value),
        len(grid) * spacing**2 * filling_factor,
        rtol=100,
    )



import numexpr as ne
import numpy as np

from pyoperators.utils import reshape_broadcast

__all__ = ['BeamGaussian', 'BeamUniformHalfSpace']


class Beam:
    def __init__(self, solid_angle):
        """
        Parameter
        ---------
        solid_angle : float
            The beam solid angle [sr].

        """
        self.solid_angle = float(solid_angle)

    def __call__(self, theta_rad, phi_rad):
        raise NotImplementedError()

    def healpix(self, nside):
        """
        Return the beam as a Healpix map.

        Parameter
        ---------
        nside : int
             The Healpix map's nside.

        """
        import healpy as hp

        npix = hp.nside2npix(nside)
        theta, phi = hp.pix2ang(nside, np.arange(npix))
        return self(theta, phi)


class BeamGaussian(Beam):
    """
    Axisymmetric gaussian beam.

    """

    def __init__(self, fwhm, backward=False):
        """
        Parameters
        ----------
        fwhm : float
            The Full-Width-Half-Maximum of the beam, in radians.
        backward : boolean, optional
            If true, the maximum of the beam is at theta=pi.

        """
        self.fwhm = fwhm
        self.sigma = fwhm / np.sqrt(8 * np.log(2))
        self.backward = bool(backward)
        Beam.__init__(self, 2 * np.pi * self.sigma**2)

    def __call__(self, theta, phi):
        if self.backward:
            theta = np.pi - theta
        coef = -0.5 / self.sigma**2
        out = ne.evaluate(
            'exp(coef * theta**2)', local_dict={'coef': coef, 'theta': theta}
        )
        return reshape_broadcast(out, np.broadcast(theta, phi).shape)


class BeamUniformHalfSpace(Beam):
    """
    Uniform beam in half-space.

    """

    def __init__(self):
        Beam.__init__(self, 2 * np.pi)

    def __call__(self, theta, phi):
        out = 1.0
        return reshape_broadcast(out, np.broadcast(theta, phi).shape)



#!/usr/bin/env python3
"""
Script to generate Fortran source files from Jinja2 templates.

This script replaces the preprocessing system previously provided by
pchanial-legacy-install-hooks for the migration to numpy 2.0 and Meson.

The old .f90.src files used a custom syntax:
    ! <variable=value1,value2,value3>

This script processes Jinja2 templates (.f90.j2) that use standard
Jinja2 syntax to generate multiple variations of Fortran code.
"""

import argparse
import re
from pathlib import Path

from jinja2 import Environment, FileSystemLoader, StrictUndefined


def parse_template_variables(template_content: str) -> tuple[dict[str, list[str]], str]:
    """
    Parse template variable definitions from the template content.

    Looks for lines like:
        {# variables: var1=val1,val2,val3; var2=val1,val2 #}

    Returns:
        A tuple of (variable_dict, cleaned_content)
        where variable_dict maps variable names to lists of values
    """
    variables = {}

    # Look for the special variables comment
    pattern = r'\{#\s*variables:\s*(.*?)\s*#\}'
    match = re.search(pattern, template_content, re.MULTILINE)

    if match:
        vars_str = match.group(1)
        # Parse each variable definition
        for var_def in vars_str.split(';'):
            var_def = var_def.strip()
            if '=' in var_def:
                name, values_str = var_def.split('=', 1)
                values = [v.strip() for v in values_str.split(',')]
                variables[name.strip()] = values

    return variables, template_content


def generate_combinations(variables: dict[str, list[str]]) -> list[dict[str, str]]:
    """
    Generate all combinations of variable values.

    All variable lists must have the same length, and combinations are
    formed by taking the i-th element from each list.

    Example:
        variables = {'ikind': ['int32', 'int64'], 'isize': ['4', '8']}
        returns: [{'ikind': 'int32', 'isize': '4'}, {'ikind': 'int64', 'isize': '8'}]
    """
    if not variables:
        return [{}]

    # Check that all lists have the same length
    lengths = [len(values) for values in variables.values()]
    if len(set(lengths)) > 1:
        raise ValueError(
            f"All variable lists must have the same length. Got: {lengths}"
        )

    if not lengths:
        return [{}]

    n = lengths[0]
    combinations = []

    for i in range(n):
        combo = {name: values[i] for name, values in variables.items()}
        combinations.append(combo)

    return combinations


def generate_fortran_file(
    template_path: Path, output_path: Path, precision_real: int = 8
) -> None:
    """
    Generate a Fortran source file from a Jinja2 template.

    Args:
        template_path: Path to the .f90.j2 template file
        output_path: Path where the generated .f90 file should be written
        precision_real: Value for PRECISION_REAL preprocessor macro (4, 8, or 16)
    """
    # Read template
    with open(template_path) as f:
        template_content = f.read()

    # Parse variable definitions
    variables, template_content = parse_template_variables(template_content)

    # Setup Jinja2 environment
    env = Environment(
        loader=FileSystemLoader(template_path.parent),
        undefined=StrictUndefined,
        trim_blocks=True,
        lstrip_blocks=True,
    )

    # Add custom filters if needed
    env.filters['fortran_type_size'] = lambda kind: {
        'int8': '1',
        'int16': '2',
        'int32': '4',
        'int64': '8',
        'real32': '4',
        'real64': '8',
        'real128': '16',
    }.get(kind, '?')

    template = env.from_string(template_content)

    # Generate combinations
    combinations = generate_combinations(variables)

    # Render template with all combinations
    # The template should use {% for combo in combinations %}...{% endfor %}
    # to generate multiple versions of functions
    output_content = template.render(
        combinations=combinations,
        PRECISION_REAL=precision_real,
        GFORTRAN=True,  # Default to gfortran
    )

    # Write output
    output_path.parent.mkdir(parents=True, exist_ok=True)
    with open(output_path, 'w') as f:
        f.write(output_content)

    print(f"Generated {output_path} from {template_path}")
    if combinations:
        print(f"  Generated {len(combinations)} function variations")


def main():
    parser = argparse.ArgumentParser(
        description='Generate Fortran source files from Jinja2 templates'
    )
    parser.add_argument(
        'templates', nargs='+', type=Path, help='Template files (.f90.j2) to process'
    )
    parser.add_argument(
        '-o',
        '--output-dir',
        type=Path,
        help='Output directory (default: same as template directory)',
    )
    parser.add_argument(
        '--precision-real',
        type=int,
        choices=[4, 8, 16],
        default=8,
        help='PRECISION_REAL value (default: 8)',
    )

    args = parser.parse_args()

    for template_path in args.templates:
        if not template_path.exists():
            print(f"Warning: Template {template_path} does not exist, skipping")
            continue

        if args.output_dir:
            output_name = template_path.name.replace('.j2', '')
            output_path = args.output_dir / output_name
        else:
            output_path = template_path.parent / template_path.name.replace('.j2', '')

        try:
            generate_fortran_file(template_path, output_path, args.precision_real)
        except Exception as e:
            print(f"Error processing {template_path}: {e}")
            raise


if __name__ == '__main__':
    main()
